################################################################
# Copyright (c) 2022-2023 Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
################################################################

from __future__ import division
from array import *
import io
import os
import sys
import re
from common import load_json

dmacs_all = []
vids_all = []
def add_del_fdb(dmac, vids, port_str, rest_str, add):
    vid_array = re.split(',',vids)
    length = len(vid_array)
    for i in range(0,length):
        vid_range = re.split('-',vid_array[i])
        l = len(vid_range)
        if l == 1:
            start = int(vid_range[0])
            end = start + 1
        else:
            start = int(vid_range[0])
            end = int(vid_range[1]) + 1
        for vid in range(start,end):
            if add == 1:
                cmd = "switch_cam2 -e "+  port_str + " -m " +  dmac + " -v " + str(vid) + rest_str
            else:
                cmd = "switch_cam2 -d -m " + dmac + " -v " + str(vid)
            os.system(cmd)
def fdb():
    num_entries = len(dmacs_all)
    if num_entries:
        num_entries = len(dmacs_all)
        for i in range(num_entries):
            add_del_fdb(dmacs_all[i],vids_all[i],"","",0)
    del dmacs_all [:]
    del vids_all [:]
    data = load_json(sys.argv[1])
    if not data:
        return
    rest_str = ""
    for bridge_ll in data["ieee802-dot1q-bridge:bridges"]["bridge"]:
        for component_ll in bridge_ll["component"]:
            for cam_ll in component_ll["ge-fdb:forwarding-table"]["forwarding-entry"]:
                dmac = cam_ll["destination-mac-address"]
                dmacs_all.append(dmac)
                port_str = ""
                rest_str = ""
                for port_ll in cam_ll["port-map"]:
                    j = 0
                    port_ref = port_ll["port-ref"]
                    for port_dest in port_ll["port-dest"]:
                        if port_dest == 3:
                            port = "--swp0"
                            if port_ref == 4:
                                rest_str = " -s swpex0"
                                swpex0 = 1
                        elif port_dest == 1:
                            port = "--swp1"
                        elif port_dest == 2:
                            port = "--swp2"
                        elif port_dest == 4:
                            port = "--swpex0"
                            if port_ref == 3:
                                rest_str = " -s swp0"	
                        else:
                            sys.exit("port-dest should contain a value between 1 and 3")
                        if "vlan-dest" in cam_ll["port-map"][0].keys():
                            vlan_dest = int(cam_ll["port-map"][0]["vlan-dest"])
                            if vlan_dest == 0:
                                port += "=u"
                        if j == 0:
                            port_str += port
                        else:
                            port_str = port_str + " " + port
                        j = j + 1
                vids = cam_ll["vids"]
                vids_all.append(vids)
                if "queue-dest" in cam_ll["port-map"][0].keys():
                            queue_dest = " -i " + str(pow(2,cam_ll["port-map"][0]["queue-dest"]))
                            rest_str += queue_dest
                add_del_fdb(dmac, vids, port_str,rest_str, 0)
                add_del_fdb(dmac, vids, port_str,rest_str, 1)

fdb()
while(1):
	cmd = 'inotifywait -e modify %s' %(sys.argv[1])
	while(os.system(cmd)):
		pass
	fdb()
    



################################################################
# Copyright (c) 2022-2023 Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
################################################################

import json
import sys
import string
import yaml
import re
import tftp_nm


def add_fdb_main(mac,fdb_file_path,self=None):
    num_entries = 0
    while (True):
        fdb_json = open(fdb_file_path, 'r+')
        data = json.load(fdb_json)
        # fdb_json.close()
        ret = add_fdb(data)
        if ret == -1:
            if num_entries > 0:
                print(str(num_entries) + " valid entries will be sent to the device")
                tftp_nm.program_device(mac,fdb_file_path)
                fdb_json.close()
            return
        else:
            num_entries = num_entries + 1
            fdb_json.close()
            fdb_json = open(fdb_file_path, "w+")
            json.dump(ret, fdb_json, indent=11)
            fdb_json.close()
            val = input(
                "Static CAM entry added locally, Want to add more entries [Y/N]? If you enter N/n, device will be "
                "programmed immediately\n")
            if val.lower() in ["y", "yes"]:
                continue
            else:
                tftp_nm.program_device(mac,fdb_file_path)
                return


def add_fdb(fdb_dict):
    mac = input("Please enter a destination MAC address for the CAM entry\n")
    if not re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", mac.lower()):
        print("Invalid MAC address\n")
        return -1
    vlan = input("Please enter a vlan ID(s)/range of vlan's for the CAM Entry\n")
    dest_ports = input(
        "Please specify destination switch port/s:\n swp0: 3, swp1: 1, swp2: 2\n Multiple ports can be "
        "specified using a comma separated string\n")
    port_list = dest_ports.split(",", 2)
    port_list_int = []
    for port in port_list:
        if int(port) not in [1, 2, 3]:
            print("Invalid name for destination switch port: " + port)
            return -1
        else:
            port_list_int.append(int(port))
    forwarding_entry = dict.fromkeys(["destination-mac-address", "vids", "port-map"])
    forwarding_entry["destination-mac-address"] = mac
    forwarding_entry["vids"] = vlan
    port_map_list = []
    port_map_dict = dict.fromkeys(["port-dest", "port-ref"])
    port_map_dict["port-ref"] = 3
    port_map_dict["port-dest"] = port_list_int
    port_map_list.append(port_map_dict)
    forwarding_entry["port-map"] = port_map_list
    fdb_dict["ieee802-dot1q-bridge:bridges"]["bridge"][0]["component"][0]["ge-fdb:forwarding-table"][
        "forwarding-entry"].append(forwarding_entry)
    return fdb_dict



import json
import time
from datetime import timedelta

def load_json(file_path):
    timeout = timedelta(seconds=1)
    sleep_time = timedelta(milliseconds=1)
    time_elapsed = timedelta(milliseconds=0)
    while time_elapsed < timeout:
        try:
            with open(file_path) as data_file:
                data = json.load(data_file)
        except json.decoder.JSONDecodeError:
            time.sleep(sleep_time.total_seconds())
            time_elapsed += sleep_time
        else:
            return data
    print(f"Time out while loading json data from '{file_path}'")
    return None



################################################################
# Copyright (c) 2022-2023 Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
################################################################

from __future__ import division
from array import *
import libconf
import io
import os
import sys
import signal
import pty
from datetime import datetime
import time
import math
from decimal import *
import shlex, subprocess
from common import load_json
DELTA = 0.1

def qbv_sched():
	data = load_json(sys.argv[1])
	if not data:
		return
	#Creating a temporary qbv config file
	f = open('qbv_cfg_bridge_tmp', 'w')
	f.write('qbv =\n{\n')
	count = 0
	error = 0
	strings = []
	first_interface = 1
	now = 0
	for int_ll in data["ietf-interfaces:interfaces"]["interface"]:
		state = array('L')
		t = array('L')
		i = 0
		gate_list_length = int_ll["ge-qbv:gate-parameters"]["admin-control-list-length"]
		cycle_time = int_ll["ge-qbv:gate-parameters"]["admin-cycle-time"]
		if cycle_time > 1000000000:
			error = 1
			print("Admin cycle time cannot be more than 1 second\n")
			break
		if int_ll["name"] == "swp0":
			strings.append("ep")
		elif int_ll["name"] == "swp1":
			strings.append(swp1_ifname)		
		else:
			strings.append(swp2_ifname)
		total_time = 0
		for conf in int_ll["ge-qbv:gate-parameters"]["admin-control-list"]:
			state.insert(i,conf["sgs-params"]["gate-states-value"]) 
			t.insert(i,conf["sgs-params"]["time-interval-value"])
			total_time = total_time + conf["sgs-params"]["time-interval-value"]
			i = i + 1
			if i == 256:
				error = 1
				print("Number of gate entries cannot be more than 256\n")
				break
			if total_time > cycle_time and cycle_time != 0:
				error = 1
				print("Sum of times of gate entries cannot be more than cycle time\n")
				break
			if conf["sgs-params"]["time-interval-value"] > 8000000:
				error = 1
				print("Time of a gate entrie cannot be more than 8ms\n")
				break
			if conf["sgs-params"]["gate-states-value"] > 7:
				error = 1
				print("gate state value cannot be greater than 7\n")
				break
		if error == 1:
			break
		string = int_ll["ge-qbv:gate-parameters"]["admin-base-time"]
		time_ns = 0
		start_sec = 0
		if(string != "0"):
			time_tuple = time.strptime(string[0:19], "%Y-%m-%dT%H:%M:%S")
			j = 0
			if string[19+j:20+j] == '.':
				j = j+1
				temp = string[19+j:20+j]
				while temp.isdigit():
					decimal = int(temp)
					time_ns = (decimal * 10**(9-j)) + time_ns
					j = j+1
					temp = string[19+j:20+j]
			json_offset = 0
			if string[19+j:20+j] != 'Z':
				json_offset_sign = 1
				if string[19+j:20+j] == '-':
					json_offset_sign = -1
				temp = string[20+j:22+j]
				json_offset = int(temp)
				temp = string[23+j:25+j]
				json_offset = json_offset_sign * ((json_offset * 60 * 60) + (int(temp) * 60))
			timestamp = time.mktime(time_tuple)
			start_sec = int(timestamp) - json_offset
		cycle_time_sec = cycle_time / (10**9)
		start_time = Decimal(start_sec) + Decimal(time_ns/(10**9))
		if (first_interface == 1):
			out_r, out_w = pty.openpty()
			process = subprocess.Popen(["ptptime",strings[-1]], stdout=out_w, preexec_fn=os.setsid)
			os.close(out_w)
			line = os.read(out_r, 1000)
			os.killpg(os.getpgid(process.pid), signal.SIGTERM)
			words = line.decode('utf8').rstrip().split(' ')
			ptptime_s = 0
			ptptime_ns = 0
			now = Decimal(ptptime_s) + Decimal(ptptime_ns/(10**9))
			first_interface = 0
		if(now>=start_time and cycle_time != 0 and start_time > 0):
			N = int ((now - start_time) / Decimal(cycle_time_sec))
			start_time = start_time + Decimal( (N+1) * cycle_time_sec)
			if ((start_time-now) < DELTA):
				N = int (DELTA / cycle_time_sec)
				start_time = start_time + Decimal( (N+1) * cycle_time_sec)
		start_sec = int(start_time)
		start_ns = int((Decimal(start_time) - Decimal(start_sec)) * (10**9))
		f.write('%s =\n{\nstart_sec = %d;\nstart_ns = %d;\ncycle_time = %d;\ngate_list_length = %d;\ngate_list =\n(\n' %(strings[count], start_sec, start_ns, cycle_time, gate_list_length))
		for k in range(0, len(state)):
			if k+1 == len(state):
				f.write('{\nstate = %d;\ntime = %d;\n}\n' % (state[k], t[k]))
			else:
				f.write('{\nstate = %d;\ntime = %d;\n},\n' % (state[k], t[k]))
		f.write(');\n};')
		count = count + 1
		if count == 3:
			break
	f.write('};')
	f.close()
	if error == 0:
		
		with io.open('qbv_cfg_bridge_tmp') as f:
			config = libconf.load(f)
		f.close()
		k = 0
		for k in range(count):
                        cmd = "qbv_sched -c %s qbv_cfg_bridge_tmp -f" %(strings[k])
                        os.system(cmd)
                        k = k + 1

proc = subprocess.Popen(["find /sys/devices/platform/ -iname  *tsn_emac_* | grep -i endpoint"],shell=True,stdout=subprocess.PIPE, stderr=subprocess.PIPE)
outs, errs = proc.communicate()
outs = outs.decode().split("\n")
swp1_ifname = ""
swp2_ifname = ""
tsn_emac_0 = ""
tsn_emac_1 = ""
for i in outs:
    if "tsn_emac_0" in i:
        tsn_emac_0 = i
    if "tsn_emac_1" in i:
        tsn_emac_1 = i
if tsn_emac_0 == "" or tsn_emac_1 == "":
    print("Cannot find temac nodes")
    exit()
swp1_ifname = os.listdir(os.path.join(tsn_emac_0 , "net"))[0]
swp2_ifname = os.listdir(os.path.join(tsn_emac_1 , "net"))[0]

qbv_sched()
while(1):
	cmd = 'inotifywait -e modify %s' %(sys.argv[1])
	while(os.system(cmd)):
		pass
	qbv_sched()

